"""
Validates the OpenAPI Specification version 3.1.1

Note that per https://spec.openapis.org/oas/v3.1.1.html#relative-references-in-api-description-uris

> URIs used as references within an OpenAPI Description, or to external documentation
> or other supplementary information such as a license, are resolved as identifiers,
> and described by this specification as URIs.

> Note that some URI fields are named url for historical reasons, but the descriptive
> text for those fields uses the correct “URI” terminology.

"""

import re
from typing import Annotated, Any, ClassVar, Self

from jsonschema.exceptions import ValidationError as JSONVSchemeValidationError
from jsonschema.protocols import Validator as JSONSchemaValidator
from jsonschema.validators import validator_for  # type: ignore
from pydantic import (
    ConfigDict,
    Discriminator,
    Field,
    RootModel,
    Tag,
    model_validator,
)

from amati import AmatiValueError
from amati import model_validators as mv
from amati._logging import Logger
from amati.fields import (
    SPDXURL,
    URI,
    SPDXIdentifier,
)
from amati.fields.commonmark import CommonMark
from amati.fields.json import JSON
from amati.fields.oas import OpenAPI
from amati.fields.spdx_licences import VALID_LICENCES
from amati.validators._discriminators import reference_object_disciminator
from amati.validators.generic import GenericObject, allow_extra_fields
from amati.validators.oas304 import (
    CallbackObject,
    ContactObject,
    EncodingObject,
    ExampleObject,
    ExternalDocumentationObject,
    HeaderObject,
    LinkObject,
    PathItemObject,
    PathsObject,
    RequestBodyObject,
    ResponseObject,
    ResponsesObject,
    ServerObject,
    TagObject,
    XMLObject,
)
from amati.validators.oas304 import SecuritySchemeObject as OAS30SecuritySchemeObject

TITLE = "OpenAPI Specification v3.1.1"

# Convenience naming to ensure that it's clear what's happening.
# https://spec.openapis.org/oas/v3.1.1.html#specification-extensions
specification_extensions = allow_extra_fields


CallbackReferenceType = Annotated[
    Annotated["CallbackObject", Tag("other")]
    | Annotated["ReferenceObject", Tag("ref")],
    Discriminator(reference_object_disciminator),
]

ExampleReferenceType = Annotated[
    Annotated["ExampleObject", Tag("other")] | Annotated["ReferenceObject", Tag("ref")],
    Discriminator(reference_object_disciminator),
]

HeaderReferenceType = Annotated[
    Annotated["HeaderObject", Tag("other")] | Annotated["ReferenceObject", Tag("ref")],
    Discriminator(reference_object_disciminator),
]

LinkReferenceType = Annotated[
    Annotated["LinkObject", Tag("other")] | Annotated["ReferenceObject", Tag("ref")],
    Discriminator(reference_object_disciminator),
]


ParameterReferenceType = Annotated[
    Annotated["ParameterObject", Tag("other")]
    | Annotated["ReferenceObject", Tag("ref")],
    Discriminator(reference_object_disciminator),
]

ResponseReferenceType = Annotated[
    Annotated["ResponseObject", Tag("other")]
    | Annotated["ReferenceObject", Tag("ref")],
    Discriminator(reference_object_disciminator),
]

RequestBodyReferenceType = Annotated[
    Annotated["RequestBodyObject", Tag("other")]
    | Annotated["ReferenceObject", Tag("ref")],
    Discriminator(reference_object_disciminator),
]

SchemaReferenceType = Annotated[
    Annotated["SchemaObject", Tag("other")] | Annotated["ReferenceObject", Tag("ref")],
    Discriminator(reference_object_disciminator),
]

SecuritySchemeReferenceType = Annotated[
    Annotated["SecuritySchemeObject", Tag("other")]
    | Annotated["ReferenceObject", Tag("ref")],
    Discriminator(reference_object_disciminator),
]


@specification_extensions("x-")
class LicenceObject(GenericObject):
    """
    A model representing the OpenAPI Specification licence object §4.8.4

    OAS uses the SPDX licence list.

    # SPECFIX: The URI is mutually exclusive of the identifier. I don't see
    the purpose of this; if the identifier is a SPDX Identifier where's the
    harm in also including the URI
    """

    name: str = Field(min_length=1)
    # What difference does Optional make here?
    identifier: SPDXIdentifier | None = None
    url: URI | None = None
    _reference_uri: ClassVar[str] = URI(
        "https://spec.openapis.org/oas/v3.1.1.html#license-object"
    )

    _not_url_and_identifier = mv.only_one_of(["url", "identifier"])

    @model_validator(mode="after")
    def check_uri_associated_with_identifier(self: Self) -> Self:
        """
        Validate that the URL matches the provided licence identifier.

        This validator checks if the URL is listed among the known URLs for the
        specified licence identifier.

        Returns:
            The validated licence object
        """
        # URI only - should warn if not SPDX
        if self.url:
            try:
                SPDXURL(self.url)
            except AmatiValueError:
                Logger.log(
                    {
                        "msg": f"{self.url} is not a valid SPDX URL",
                        "type": "warning",
                        "loc": (self.__class__.__name__,),
                        "input": self.url,
                        "url": self._reference_uri,
                    }
                )

        # Both Identifier and URI, technically invalid, but should check if
        # consistent
        if (
            self.url
            and self.identifier
            and str(self.url) not in VALID_LICENCES[self.identifier]
        ):
            Logger.log(
                {
                    "msg": f"{self.url} is not associated with the identifier {self.identifier}",  # noqa: E501
                    "type": "warning",
                    "loc": (self.__class__.__name__,),
                    "input": self.model_dump_json(),
                    "url": self._reference_uri,
                }
            )

        return self


class ReferenceObject(GenericObject):
    """
    Validates the OpenAPI Specification reference object - §4.8.23

    Note, "URIs" can be prefixed with a hash; this is because if the
    representation of the referenced document is JSON or YAML, then
    the fragment identifier SHOULD be interpreted as a JSON-Pointer
    as per RFC6901.
    """

    model_config: ClassVar[ConfigDict] = ConfigDict(
        extra="forbid", populate_by_name=True
    )

    ref: URI = Field(alias="$ref")
    summary: str | None
    description: CommonMark | None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.1.1.html#reference-object"
    )


@specification_extensions("x-")
class InfoObject(GenericObject):
    """
    Validates the OpenAPI Specification info object - §4.8.2:
    """

    title: str
    summary: str | None = None
    description: str | CommonMark | None = None
    termsOfService: str | None = None
    contact: ContactObject | None = None
    license: LicenceObject | None = None
    version: str
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/3.1.1.html#info-object"
    )


@specification_extensions("x-")
class DiscriminatorObject(GenericObject):
    """
    Validates the OpenAPI Specification object - §4.8.25
    """

    # FIXME: Need post processing to determine whether the property actually exists
    # FIXME: The component and schema objects need to check that this is being used
    # properly.
    propertyName: str
    mapping: dict[str, str | URI] | None = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.1.1.html#discriminator-object"
    )


@specification_extensions("x-")
class ServerVariableObject(GenericObject):
    """
    Validates the OpenAPI Specification server variable object - §4.8.6
    """

    enum: list[str] | None = Field(None, min_length=1)
    default: str = Field(min_length=1)
    description: str | CommonMark | None = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.1.1.html#server-variable-object"
    )

    @model_validator(mode="after")
    def check_enum_default(self: Self) -> Self:
        """
        Validate that the default value is in the enum list.

        Returns:
            The validated server variable object
        """
        if self.enum is None:
            return self

        if self.default not in self.enum:
            Logger.log(
                {
                    "msg": f"The default value {self.default} is not in the enum list {self.enum}",  # noqa: E501
                    "type": "value_error",
                    "loc": (self.__class__.__name__,),
                    "input": {"default": self.default, "enum": self.enum},
                    "url": self._reference_uri,
                }
            )

        return self


PARAMETER_STYLES: set[str] = {
    "matrix",
    "label",
    "simple",
    "form",
    "spaceDelimited",
    "pipeDelimited",
    "deepObject",
}


@specification_extensions("x-")
class ParameterObject(GenericObject):
    """Validates the OpenAPI Specification parameter object - §4.8.11"""

    name: str
    in_: str = Field(alias="in")
    description: str | CommonMark | None = None
    required: bool | None = None
    deprecated: bool | None = None
    allowEmptyValue: bool | None = None
    style: str | None = None
    explode: bool | None = None
    allowReserved: bool | None = None
    schema_: "SchemaObject | None" = Field(alias="schema")  # noqa: UP037
    example: Any | None = None
    examples: dict[str, ExampleReferenceType] | None = None
    content: dict[str, "MediaTypeObject"] | None = None  # noqa: UP037
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.1.1.html#parameter-object"
    )

    _in_valid = mv.if_then(
        conditions={"in_": mv.UNKNOWN},
        consequences={"in_": ["query", "header", "path", "cookie"]},
    )
    _path_location_is_required = mv.if_then(
        conditions={"in_": "path"}, consequences={"required": True}
    )
    _empty_value_only_with_query = mv.if_then(
        conditions={"allowEmptyValue": mv.UNKNOWN},
        consequences={"in_": PARAMETER_STYLES ^ {"query"}},
    )
    _style_is_valid = mv.if_then(
        conditions={"style": mv.UNKNOWN}, consequences={"style": list(PARAMETER_STYLES)}
    )
    _reserved_only_with_query = mv.if_then(
        conditions={"allowReserved": mv.UNKNOWN},
        consequences={"in_": PARAMETER_STYLES ^ {"query"}},
    )
    _disallowed_if_schema = mv.if_then(
        conditions={"schema_": mv.UNKNOWN}, consequences={"content": None}
    )
    _disallowed_if_content = mv.if_then(
        conditions={"content": mv.UNKNOWN},
        consequences={
            "content": None,
            "style": None,
            "explode": None,
            "allowReserved": None,
            "schema_": None,
        },
    )


@specification_extensions("x-")
class MediaTypeObject(GenericObject):
    """
    Validates the OpenAPI Specification media type object - §4.8.14
    """

    schema_: "SchemaObject | None" = Field(alias="schema", default=None)  # noqa: UP037
    # FIXME: Define example
    example: Any | None = None
    examples: dict[str, ExampleReferenceType] | None = None
    encoding: EncodingObject | None = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.1.1.html#media-type-object"
    )


class SchemaObject(GenericObject):
    """
    Schema Object as per OAS 3.1.1 specification (section 4.8.24)

    This model defines only the OpenAPI-specific fields explicitly.
    Standard JSON Schema fields are allowed via the 'extra' config
    and validated through jsonschema.
    """

    model_config: ClassVar[ConfigDict] = ConfigDict(
        populate_by_name=True,
        extra="allow",  # Allow all standard JSON Schema fields
    )

    # OpenAPI-specific fields not in standard JSON Schema
    nullable: bool | None = None  # OAS 3.0 style nullable flag
    discriminator: DiscriminatorObject | None = None  # Polymorphism support
    readOnly: bool | None = None  # Declares property as read-only for requests
    writeOnly: bool | None = None  # Declares property as write-only for responses
    xml: XMLObject | None = None  # XML metadata
    externalDocs: ExternalDocumentationObject | None = None  # External documentation
    example: Any | None = None  # Example of schema
    examples: list[Any] | None = None  # Examples of schema (OAS 3.1)
    deprecated: bool | None = None  # Specifies schema is deprecated

    # JSON Schema fields that need special handling in OAS context
    ref: str | None = Field(default=None, alias="$ref")  # Reference to another schema

    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.1.1.html#schema-object"
    )

    @model_validator(mode="after")
    def validate_schema(self):
        """
        Use jsonschema to validate the model as a valid JSON Schema
        """
        schema_dict = self.model_dump(exclude_none=True, by_alias=True)

        # Handle OAS 3.1 specific validations

        # 1. Convert nullable to type array with null if needed
        if schema_dict.get("nullable") is True and "type" in schema_dict:
            type_val = schema_dict["type"]
            if isinstance(type_val, str) and type_val != "null":
                schema_dict["type"] = [type_val, "null"]
            elif isinstance(type_val, list) and "null" not in type_val:
                schema_dict["type"] = [*type_val, ["null"]]

        # 2. Validate the schema structure using jsonschema's meta-schema
        # Get the right validator based on the declared $schema or default
        # to Draft 2020-12
        schema_version = schema_dict.get(
            "$schema", "https://json-schema.org/draft/2020-12/schema"
        )
        try:
            validator_cls: JSONSchemaValidator = validator_for(  # type: ignore
                {"$schema": schema_version}
            )
            meta_schema: JSON = validator_cls.META_SCHEMA  # type: ignore

            # This will validate the structure conforms to JSON Schema
            validator_cls(meta_schema).validate(schema_dict)  # type: ignore
        except JSONVSchemeValidationError as e:
            Logger.log(
                {
                    "msg": f"Invalid JSON Schema: {e.message}",
                    "type": "value_error",
                    "loc": (self.__class__.__name__,),
                    "input": schema_dict,
                    "url": self._reference_uri,
                }
            )

        return self


class SecuritySchemeObject(OAS30SecuritySchemeObject):
    """
    Validates the OpenAPI Security Scheme object - §4.8.27
    """

    _SECURITY_SCHEME_TYPES: ClassVar[set[str]] = {
        "apiKey",
        "http",
        "oauth2",
        "openIdConnect",
        "mutualTLS",
    }


type _Requirement = dict[str, list[str]]


# NB This is implemented as a RootModel as there are no pre-defined field names.
class SecurityRequirementObject(RootModel[list[_Requirement] | _Requirement]):
    """
    Validates the OpenAPI Specification security requirement object - §4.8.30:
    """

    # FIXME: The name must be a valid Security Scheme - need to use post-processing
    # FIXME If the security scheme is of type "oauth2" or "openIdConnect", then the
    # value must be a list
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/3.1.1.html#security-requirement-object"
    )


@specification_extensions("x-")
class OperationObject(GenericObject):
    """Validates the OpenAPI Specification operation object - §4.8.10"""

    tags: list[str] | None = None
    summary: str | None = None
    description: str | CommonMark | None = None
    externalDocs: ExternalDocumentationObject | None = None
    operationId: str | None = None
    parameters: list[ParameterReferenceType] | None = None
    requestBody: RequestBodyReferenceType | None = None
    responses: ResponsesObject | None = None
    callbacks: dict[str, CallbackReferenceType] | None = None
    deprecated: bool | None = False
    security: list[SecurityRequirementObject] | None = None
    servers: list[ServerObject] | None = None

    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.1.1.html#operation-object"
    )


@specification_extensions("x-")
class ComponentsObject(GenericObject):
    """
    Validates the OpenAPI Specification components object - §4.8.7
    """

    schemas: dict[str, SchemaReferenceType] | None = None
    responses: dict[str, ResponseReferenceType] | None = None
    parameters: dict[str, ParameterReferenceType] | None = None
    examples: dict[str, ExampleReferenceType] | None = None
    requestBodies: dict[str, RequestBodyReferenceType] | None = None
    headers: dict[str, HeaderReferenceType] | None = None
    securitySchemes: dict[str, SecuritySchemeReferenceType] | None = None
    links: dict[str, LinkReferenceType] | None = None
    callbacks: dict[str, CallbackReferenceType] | None = None
    pathItems: dict[str, PathItemObject] | None = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.1.1.html#components-object"
    )

    @model_validator(mode="before")
    @classmethod
    def validate_all_fields(
        cls, data: dict[str, dict[str, Any]]
    ) -> dict[str, dict[str, Any]]:
        """
        Validates the components object.

        Args:
            data: The data to validate.

        Returns:
            The validated components object.
        """

        pattern: str = r"^[a-zA-Z0-9\.\-_]+$"

        # Validate each field in the components object
        for field_name, value in data.items():
            if field_name.startswith("x-"):
                continue

            if not isinstance(value, dict):  # type: ignore
                raise ValueError(
                    f"Invalid type for '{field_name}': expected dict, got {type(value)}"
                )

            for key in value:
                if not re.match(pattern, key):
                    raise ValueError(
                        f"Invalid key '{key}' in '{field_name}': must match pattern {pattern}"  # noqa: E501
                    )

        return data


@specification_extensions("x-")
class OpenAPIObject(GenericObject):
    """
    Validates the OpenAPI Specification object - §4.1
    """

    openapi: OpenAPI
    info: InfoObject
    jsonSchemaDialect: URI | None = None
    servers: list[ServerObject] | None = Field(default=[ServerObject(url=URI("/"))])
    paths: PathsObject | None = None
    webhooks: dict[str, PathItemObject] | None = None
    components: ComponentsObject | None = None
    security: list[SecurityRequirementObject] | None = None
    tags: list[TagObject] | None = None
    externalDocs: ExternalDocumentationObject | None = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/3.1.1.html#openapi-object"
    )



import importlib
import json
import time
from pathlib import Path
from typing import Literal

from loguru import logger
from requests.exceptions import HTTPError

data_types = Literal[
    "http_status_code",
    "iso9110",
    "media_types",
    "schemes",
    "spdx_licences",
    "tlds",
    "all",
]

type JSONPrimitive = str | int | float | bool | None
type JSONArray = list["JSONValue"]
type JSONObject = dict[str, "JSONValue"]
type JSONValue = JSONPrimitive | JSONArray | JSONObject


def refresh_module(module_: data_types) -> None:
    module = importlib.import_module(f"amati._data.{module_}")

    if not hasattr(module, "get"):
        raise ImportError(f"{module_} does not have a 'get' function.")

    if not hasattr(module, "DATA_FILE"):
        raise ImportError(f"{module_} does not have a 'DATA_FILE' attribute")

    if not hasattr(module, "DATA_SOURCE"):
        raise ImportError(f"{module_} does not have a 'DATA_SOURCE' attribute")

    logger.info(f"Refreshing data from {module_}")

    try:
        data = module.get()
    except HTTPError as e:
        logger.error(
            f"Failed to fetch data from {module.DATA_SOURCE}: {e}. "
            "Please check your internet connection or the availability of the data."
        )
        return

    if data is None:
        raise ValueError(
            f"Get {module_} returned None. Please check whether "
            f"{module.DATA_SOURCE} remains available."
        )

    current_path = Path(__file__).parent / "files"
    write_to = current_path / module.DATA_FILE

    with write_to.open("w") as f:
        json.dump(data, f, indent=2)


def _handle_data_types(data_type: data_types | list[data_types]) -> list[data_types]:
    """
    Double checks that the presented data_type is valid.

    Args:
        data_type: A string or a list of strings representing the data types to refresh.

    Returns:
        A list of valid data types.

    Raises:
        TypeError: If data_type is not a string or a list of strings.
        ValueError: If any of the provided data types are not valid.

    >>> _handle_data_types("http_status_code")
    ['http_status_code']
    >>> _handle_data_types(["http_status_code", "media_types"])
    ['http_status_code', 'media_types']
    >>> _handle_data_types("all")
    ['http_status_code', 'iso9110', 'media_types', 'schemes', 'spdx_licences', 'tlds']
    """

    if isinstance(data_type, str):
        if data_type == "all":
            data_type = [x for x in data_types.__args__ if x != "all"]
        else:
            data_type = [data_type]

    if not isinstance(data_type, list):  # type: ignore
        raise TypeError("data_type must be a string or a list of strings.")

    for module in data_type:
        if module not in data_types.__args__:
            raise ValueError(
                f"Invalid data type: {module}. Must be one of {data_types}."
            )

    return data_type


def refresh(data_type: data_types | list[data_types]) -> None:
    """
    Refreshes the data for the specified data type(s).
    Args:
        data_type: A string or a list of strings representing the data types to refresh.
            Can be one of:
            - "http_status_code"
            - "iso9110"
            - "media_types"
            - "schemes"
            - "spdx_licences"
            - "tlds"
            - "all" (to refresh all data types)
    Returns:
        None
    Raises:
        TypeError: If data_type is not a string or a list of strings.
        ValueError: If any of the provided data types are not valid.
    """
    to_refresh = _handle_data_types(data_type)

    for module in to_refresh:
        refresh_module(module)

        time.sleep(1)  # To avoid hitting IANA too hard


def get(data_type: data_types) -> dict[str, JSONValue] | list[JSONValue] | None:
    """
    Returns the data for a specified data type.

    Args:
        data_type: A string representing the data type to retrieve. Can be one of:
            - "http_status_code"
            - "iso9110"
            - "media_types"
            - "schemes"
            - "spdx_licences"
            - "tlds"

    Returns:
        A generator yielding JSON data for the specified data type.
    Raises:
        TypeError: If data_type is not a string.
        ValueError: If data_type is not one of the valid types.

    """
    to_get = _handle_data_types(data_type)

    if len(to_get) != 1:
        raise ValueError("get() can only retrieve one data type at a time. ")

    data = None
    current_path = Path(__file__).parent / "files"

    for module_ in to_get:
        module = importlib.import_module(f"amati._data.{module_}")

        data_file: Path = current_path / module.DATA_FILE

        with data_file.open(encoding="utf-8") as f:
            data = json.loads(f.read())

        if not data:
            logger.error(
                f"No data found for {module_}. "
                "Please ensure the data has been refreshed."
            )

    return data



"""
Validates the idenfitier and licences from the System Package Data
Exchange (SPDX) licence list.
"""

from typing import Any, cast

from amati import get
from amati.exceptions import AmatiValueError
from amati.fields import Str as _Str
from amati.fields.uri import URI

reference_uri = "https://spdx.org/licenses/"

data: list[dict[str, Any]] = cast(list[dict[str, Any]], get("spdx_licences"))

# `seeAlso` is the list of URLs associated with each licence
VALID_LICENCES: dict[str, list[str]] = {
    licence["licenseId"]: licence["seeAlso"] for licence in data
}
VALID_URLS: list[str] = [url for urls in VALID_LICENCES.values() for url in urls]


class SPDXIdentifier(_Str):
    """
    A class representing a valid SPDX license identifier.

    This class validates that a string value is a registered SPDX license identifier
    according to the official SPDX license list. It inherits from _Str to maintain
    compatibility with string operations while adding SPDX-specific validation.

    SPDX identifiers are standardized short-form identifiers for open source licenses,
    such as "MIT", "Apache-2.0", or "GPL-3.0-only".

    Attributes:
        Inherits all attributes from _Str

    Example:
        >>> license_id = SPDXIdentifier("MIT")
        >>> str(license_id)
        'MIT'
        >>> SPDXIdentifier("InvalidLicense")
        Traceback (most recent call last):
        amati.AmatiValueError: message
    """

    def __init__(self, value: str):
        if value not in VALID_LICENCES:
            raise AmatiValueError(
                f"{value} is not a valid SPDX licence identifier", reference_uri
            )


class SPDXURL(URI):
    """
    A class representing a valid SPDX license URL.

    This class validates that a URI is associated with an SPDX license in the official
    SPDX license list. It inherits from URI to maintain compatibility with URI
    validation and operations while adding SPDX-specific validation.

    SPDX license URLs are the official reference URLs for licenses in the SPDX registry,
    typically pointing to the canonical text of the license.

    Attributes:
        Inherits all attributes from URI

    Example:
        >>> license_url = SPDXURL("https://www.apache.org/licenses/LICENSE-2.0")
        >>> license_url.scheme
        'https'
        >>> SPDXURL("https://example.com")
        Traceback (most recent call last):
        amati.AmatiValueError: message
    """

    def __init__(self, value: str):
        super().__init__(value)

        if value not in VALID_URLS:
            raise AmatiValueError(
                f"{value} is not associated with any identifier.", reference_uri
            )



from pathlib import Path

import pytest
from hypothesis import assume, given
from hypothesis import strategies as st
from pydantic import BaseModel

from amati._references import URIReference
from amati.fields import URI, URIType
from tests.strategies import (
    absolute_paths,
    json_pointers,
    network_path_uris,
    pydantic_models,
    relative_paths,
    relative_uris,
)


class TestURIReferenceResolveExamples:
    """Example-based tests for URIReference.resolve() with concrete scenarios"""

    def test_relative_uri_simple_sibling(self) -> None:
        """Relative URI to sibling file resolves correctly"""
        uri_ref: URIReference = URIReference(
            uri=URI("sibling.json"),
            source_document=Path("/project/schemas/main.json"),
            source_model_name="MainModel",
            source_field="ref_field",
            target_model=BaseModel,
        )

        assert uri_ref.resolve() == Path("/project/schemas/sibling.json")

    def test_relative_uri_parent_directory(self) -> None:
        """Relative URI with parent traversal resolves correctly"""
        uri_ref: URIReference = URIReference(
            uri=URI("../sibling.json"),
            source_document=Path("/project/schemas/main.json"),
            source_model_name="MainModel",
            source_field="ref_field",
            target_model=BaseModel,
        )

        assert uri_ref.resolve() == Path("/project/sibling.json")

    def test_relative_uri_subdirectory(self) -> None:
        """Relative URI to subdirectory resolves correctly"""
        uri_ref: URIReference = URIReference(
            uri=URI("components/types.json"),
            source_document=Path("/project/schemas/main.json"),
            source_model_name="MainModel",
            source_field="ref_field",
            target_model=BaseModel,
        )

        assert uri_ref.resolve() == Path("/project/schemas/components/types.json")

    def test_relative_uri_multiple_parent_traversal(self) -> None:
        """Relative URI with multiple parent traversals resolves correctly"""
        uri_ref: URIReference = URIReference(
            uri=URI("../../other/file.json"),
            source_document=Path("/a/b/c/doc.json"),
            source_model_name="MainModel",
            source_field="ref_field",
            target_model=BaseModel,
        )

        assert uri_ref.resolve() == Path("/a/other/file.json")

    def test_relative_uri_with_dot_current_dir(self) -> None:
        """Relative URI with ./current directory notation resolves correctly"""
        uri_ref: URIReference = URIReference(
            uri=URI("./sibling.json"),
            source_document=Path("/project/schemas/main.json"),
            source_model_name="MainModel",
            source_field="ref_field",
            target_model=BaseModel,
        )

        assert uri_ref.resolve() == Path("/project/schemas/sibling.json")

    def test_absolute_file_path_returns_itself(self) -> None:
        """Absolute file path URIs resolve to themselves"""
        uri_ref: URIReference = URIReference(
            uri=URI("/absolute/path/file.json"),
            source_document=Path("/project/schemas/main.json"),
            source_model_name="MainModel",
            source_field="ref_field",
            target_model=BaseModel,
        )

        assert uri_ref.resolve() == Path("/project/schemas/absolute/path/file.json")

    def test_uri_with_http_scheme_raises_not_implemented(self) -> None:
        """HTTP URI with scheme should not be resolved relative to source"""
        uri_ref: URIReference = URIReference(
            uri=URI("http://example.com/schema.json"),
            source_document=Path("/project/schemas/main.json"),
            source_model_name="MainModel",
            source_field="ref_field",
            target_model=BaseModel,
        )

        with pytest.raises(NotImplementedError):
            uri_ref.resolve()

    def test_uri_with_file_scheme_not_resolved_relatively(self) -> None:
        """File URI with scheme should not be resolved relative to source"""
        uri_ref: URIReference = URIReference(
            uri=URI("file:///absolute/path/schema.json"),
            source_document=Path("/project/schemas/main.json"),
            source_model_name="MainModel",
            source_field="ref_field",
            target_model=BaseModel,
        )

        resolved: Path = uri_ref.resolve()

        assert resolved == Path("/absolute/path/schema.json"), (
            "Absolute URI resolved as whole path"
        )
        assert resolved != Path("/project/schemas/file:///absolute/path/schema.json")

    def test_network_path_uri_resolves_to_itself(self) -> None:
        """Network path URIs (//host/path) resolve to themselves"""
        uri_ref: URIReference = URIReference(
            uri=URI("//example.com/path/schema.json"),
            source_document=Path("/project/schemas/main.json"),
            source_model_name="MainModel",
            source_field="ref_field",
            target_model=BaseModel,
        )

        resolved: Path = uri_ref.resolve()

        # Network paths are absolute per RFC 3987
        assert resolved == Path("/example.com/path/schema.json")

    def test_pathless_uri_raises_value_error(self) -> None:
        """File URI without a path component should raise ValueError"""
        uri_ref: URIReference = URIReference(
            uri=URI("file://a.com"),
            source_document=Path("/project/schemas/main.json"),
            source_model_name="MainModel",
            source_field="ref_field",
            target_model=BaseModel,
        )

        with pytest.raises(ValueError):
            uri_ref.resolve()


class TestURIReferenceResolveProperties:
    """Property-based tests for URIReference.resolve() invariants"""

    @given(
        uri=st.one_of(relative_uris(), network_path_uris()),
        source_document=absolute_paths(),
        source_model_name=st.text(min_size=1),
        source_field=st.text(min_size=1),
        target_model=pydantic_models(),
    )
    def test_resolve_is_idempotent(
        self,
        uri: URI,
        source_document: Path,
        source_model_name: str,
        source_field: str,
        target_model: type[BaseModel],
    ) -> None:
        """Calling resolve() multiple times should return the same result"""
        uri_ref: URIReference = URIReference(
            uri=uri,
            source_document=source_document,
            source_model_name=source_model_name,
            source_field=source_field,
            target_model=target_model,
        )

        first_resolve: Path = uri_ref.resolve()
        second_resolve: Path = uri_ref.resolve()
        third_resolve: Path = uri_ref.resolve()

        assert first_resolve == second_resolve == third_resolve

    @given(
        uri=st.one_of(relative_uris(), network_path_uris(), json_pointers()),
        source_document=st.one_of(absolute_paths(), relative_paths()),
        source_model_name=st.text(min_size=1),
        source_field=st.text(min_size=1),
        target_model=pydantic_models(),
    )
    def test_resolve_always_returns_path(
        self,
        uri: URI,
        source_document: Path,
        source_model_name: str,
        source_field: str,
        target_model: type[BaseModel],
    ) -> None:
        """resolve() should always return a Path object"""

        assume(not (uri.type == URIType.NETWORK_PATH and source_document.is_absolute()))
        uri_ref: URIReference = URIReference(
            uri=uri,
            source_document=source_document,
            source_model_name=source_model_name,
            source_field=source_field,
            target_model=target_model,
        )

        resolved: Path = uri_ref.resolve()

        assert isinstance(resolved, Path)

    @given(
        relative_uri=relative_uris(),
        source_document=st.one_of(absolute_paths(), relative_paths()),
        source_model_name=st.text(min_size=1),
        source_field=st.text(min_size=1),
        target_model=pydantic_models(),
    )
    def test_relative_uri_path_is_normalized(
        self,
        relative_uri: URI,
        source_document: Path,
        source_model_name: str,
        source_field: str,
        target_model: type[BaseModel],
    ) -> None:
        """Relative URI paths should be normalized (.. and . resolved)"""
        assume(relative_uri.type == URIType.RELATIVE)
        assume(relative_uri.scheme is None)

        uri_ref: URIReference = URIReference(
            uri=relative_uri,
            source_document=source_document,
            source_model_name=source_model_name,
            source_field=source_field,
            target_model=target_model,
        )

        resolved: Path = uri_ref.resolve()

        assert resolved.is_absolute(), "Resolved path should be absolute (normalized)"

        assert ".." not in resolved.parts
        assert "." not in resolved.parts

    @given(
        relative_uri=relative_uris(),
        source_document=absolute_paths(),
        source_model_name=st.text(min_size=1),
        source_field=st.text(min_size=1),
        target_model=pydantic_models(),
    )
    def test_relative_uri_not_resolved_from_cwd(
        self,
        relative_uri: URI,
        source_document: Path,
        source_model_name: str,
        source_field: str,
        target_model: type[BaseModel],
    ) -> None:
        """Relative URIs should NOT resolve from current working directory"""
        assume(relative_uri.type == URIType.RELATIVE)
        assume(relative_uri.scheme is None)
        assume(source_document.parent != Path.cwd())
        assume(relative_uri.path != "/")

        uri_ref: URIReference = URIReference(
            uri=relative_uri,
            source_document=source_document,
            source_model_name=source_model_name,
            source_field=source_field,
            target_model=target_model,
        )

        resolved: Path = uri_ref.resolve()
        cwd_resolved: Path = (Path.cwd() / relative_uri.lstrip("/")).resolve()

        if source_document.parent != Path.cwd():
            assert resolved != cwd_resolved, "Should be different if source not in cwd"

    @given(
        relative_uri=relative_uris(),
        source_document=absolute_paths(),
        source_model_name=st.text(min_size=1),
        source_field=st.text(min_size=1),
        target_model=pydantic_models(),
    )
    def test_relative_uri_not_resolved_from_source_document_itself(
        self,
        relative_uri: URI,
        source_document: Path,
        source_model_name: str,
        source_field: str,
        target_model: type[BaseModel],
    ) -> None:
        """Relative URIs should resolve from parent directory,
        not the document itself"""
        assume(relative_uri.type == URIType.RELATIVE)
        assume(relative_uri.scheme is None)
        assume(relative_uri.path != "/")

        uri_ref: URIReference = URIReference(
            uri=relative_uri,
            source_document=source_document,
            source_model_name=source_model_name,
            source_field=source_field,
            target_model=target_model,
        )

        resolved: Path = uri_ref.resolve()
        wrong_from_document: Path = (
            source_document / relative_uri.lstrip("/")
        ).resolve()

        if source_document.parent != source_document:
            assert resolved != wrong_from_document, (
                "Should NOT resolve from the document itself"
            )



"""
Validates the OpenAPI Specification version 3.1.1

Note that per https://spec.openapis.org/oas/v3.0.4.html#relative-references-in-api-description-uris

> URIs used as references within an OpenAPI Description, or to external documentation
> or other supplementary information such as a license, are resolved as identifiers,
> and described by this specification as URIs.

> Note that some URI fields are named url for historical reasons, but the descriptive
> text for those fields uses the correct “URI” terminology.

"""

import re
from typing import Annotated, Any, ClassVar, Self

from jsonschema.exceptions import ValidationError as JSONVSchemeValidationError
from jsonschema.protocols import Validator as JSONSchemaValidator
from jsonschema.validators import validator_for  # type: ignore
from pydantic import (
    ConfigDict,
    Discriminator,
    Field,
    RootModel,
    SerializerFunctionWrapHandler,
    Tag,
    ValidationError,
    field_validator,
    model_serializer,
    model_validator,
)

from amati import AmatiValueError
from amati import model_validators as mv
from amati._logging import Logger
from amati.fields import (
    URI,
    Email,
    HTTPAuthenticationScheme,
    HTTPStatusCode,
    MediaType,
    URIType,
    URIWithVariables,
)
from amati.fields.commonmark import CommonMark
from amati.fields.json import JSON
from amati.fields.oas import OpenAPI, RuntimeExpression
from amati.validators._discriminators import reference_object_disciminator
from amati.validators.generic import GenericObject, allow_extra_fields

type JSONPrimitive = str | int | float | bool | None
type JSONArray = list["JSONValue"]
type JSONObject = dict[str, "JSONValue"]
type JSONValue = JSONPrimitive | JSONArray | JSONObject


TITLE = "OpenAPI Specification v3.0.4"

# Convenience naming to ensure that it's clear what's happening.
# https://spec.openapis.org/oas/v3.0.4.html#specification-extensions
specification_extensions = allow_extra_fields


CallbackReferenceType = Annotated[
    Annotated["CallbackObject", Tag("other")]
    | Annotated["ReferenceObject", Tag("ref")],
    Discriminator(reference_object_disciminator),
]

ExampleReferenceType = Annotated[
    Annotated["ExampleObject", Tag("other")] | Annotated["ReferenceObject", Tag("ref")],
    Discriminator(reference_object_disciminator),
]

HeaderReferenceType = Annotated[
    Annotated["HeaderObject", Tag("other")] | Annotated["ReferenceObject", Tag("ref")],
    Discriminator(reference_object_disciminator),
]

LinkReferenceType = Annotated[
    Annotated["LinkObject", Tag("other")] | Annotated["ReferenceObject", Tag("ref")],
    Discriminator(reference_object_disciminator),
]


ParameterReferenceType = Annotated[
    Annotated["ParameterObject", Tag("other")]
    | Annotated["ReferenceObject", Tag("ref")],
    Discriminator(reference_object_disciminator),
]

ResponseReferenceType = Annotated[
    Annotated["ResponseObject", Tag("other")]
    | Annotated["ReferenceObject", Tag("ref")],
    Discriminator(reference_object_disciminator),
]

RequestBodyReferenceType = Annotated[
    Annotated["RequestBodyObject", Tag("other")]
    | Annotated["ReferenceObject", Tag("ref")],
    Discriminator(reference_object_disciminator),
]

SchemaReferenceType = Annotated[
    Annotated["SchemaObject", Tag("other")] | Annotated["ReferenceObject", Tag("ref")],
    Discriminator(reference_object_disciminator),
]

SecuritySchemeReferenceType = Annotated[
    Annotated["SecuritySchemeObject", Tag("other")]
    | Annotated["ReferenceObject", Tag("ref")],
    Discriminator(reference_object_disciminator),
]


@specification_extensions("x-")
class ContactObject(GenericObject):
    """
    Validates the OpenAPI Specification contact object - §4.8.3
    """

    name: str | None = None
    url: URI | None = None
    email: Email | None = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/3.0.4.html#contact-object"
    )


@specification_extensions("x-")
class LicenceObject(GenericObject):
    """
    A model representing the OpenAPI Specification licence object §4.8.4
    """

    name: str = Field(min_length=1)
    url: URI | None = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#license-object"
    )


class ReferenceObject(GenericObject):
    """
    Validates the OpenAPI Specification reference object - §4.8.23

    Note, "URIs" can be prefixed with a hash; this is because if the
    representation of the referenced document is JSON or YAML, then
    the fragment identifier SHOULD be interpreted as a JSON-Pointer
    as per RFC6901.
    """

    model_config: ClassVar[ConfigDict] = ConfigDict(
        extra="forbid", populate_by_name=True
    )

    ref: URI = Field(alias="$ref")
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#reference-object"
    )


@specification_extensions("x-")
class InfoObject(GenericObject):
    """
    Validates the OpenAPI Specification info object - §4.8.2:
    """

    title: str
    description: str | CommonMark | None = None
    termsOfService: str | None = None
    contact: ContactObject | None = None
    license: LicenceObject | None = None
    version: str
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/3.0.4.html#info-object"
    )


class DiscriminatorObject(GenericObject):
    """
    Validates the OpenAPI Specification object - §4.8.25
    """

    # FIXME: Need post processing to determine whether the property actually exists
    # FIXME: The component and schema objects need to check that this is being used
    # properly.
    propertyName: str
    mapping: dict[str, str | URI] | None = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#discriminator-object"
    )


@specification_extensions("x-")
class ExampleObject(GenericObject):
    """
    Validates the OpenAPI Specification example object - §4.8.19
    """

    summary: str | None = None
    description: str | CommonMark | None = None
    value: JSONValue | None = None
    externalValue: URI | None = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#example-object"
    )

    # Note https://spec.openapis.org/oas/v3.1.1.html#example-object doesn't
    # state that one of value and externalValue is required, however,
    # there's very little point to an example object without either.
    _not_value_and_external_value = mv.only_one_of(
        ["value", "externalValue"], "warning"
    )


@specification_extensions("x-")
class ServerVariableObject(GenericObject):
    """
    Validates the OpenAPI Specification server variable object - §4.8.6
    """

    enum: list[str] | None = Field(None, min_length=1)
    default: str = Field(min_length=1)
    description: str | CommonMark | None = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#server-variable-object"
    )

    @model_validator(mode="after")
    def check_enum_default(self: Self) -> Self:
        """
        Validate that the default value is in the enum list.

        Returns:
            The validated server variable object
        """
        if self.enum is None:
            return self

        if self.default not in self.enum:
            Logger.log(
                {
                    "msg": f"The default value {self.default} is not in the enum list {self.enum}",  # noqa: E501
                    "type": "warning",
                    "loc": (self.__class__.__name__,),
                    "input": {"default": self.default, "enum": self.enum},
                    "url": self._reference_uri,
                }
            )

        return self


@specification_extensions("x-")
class ServerObject(GenericObject):
    """
    Validates the OpenAPI Specification server object - §4.8.5
    """

    url: URIWithVariables | URI
    description: str | CommonMark | None = None
    variables: dict[str, ServerVariableObject] | None = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#server-object"
    )


@specification_extensions("x-")
class ExternalDocumentationObject(GenericObject):
    """
    Validates the OpenAPI Specification external documentation object - §4.8.22
    """

    description: str | CommonMark | None = None
    url: URI
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#external-documentation-object"
    )


# FIXME: Specification extensions should be "^x-", but the implementation
# doesn't play well with ConfigDict(extra="allow"). This is the only case
# so less important to change as the eventual logic is still correct.
@specification_extensions(".*")
class PathsObject(GenericObject):
    """Validates the OpenAPI Specification paths object - §4.8.8"""

    model_config: ClassVar[ConfigDict] = ConfigDict(extra="allow")

    @model_validator(mode="before")
    @classmethod
    def paths_are_uris(cls, data: Any) -> Any:
        """
        Validates that paths are valid URIs, it's allowed that they
        have variables, e.g. /pets or /pets/{petID}

        Special-case specification extensions, which are also allowed.
        """

        for field in data:
            # Specification extensions
            if field.startswith("x-"):
                continue

            URIWithVariables(field)

        return data


PARAMETER_STYLES: set[str] = {
    "matrix",
    "label",
    "simple",
    "form",
    "spaceDelimited",
    "pipeDelimited",
    "deepObject",
}


@specification_extensions("x-")
class ParameterObject(GenericObject):
    """Validates the OpenAPI Specification parameter object - §4.8.11"""

    name: str
    in_: str = Field(alias="in")
    description: str | CommonMark | None = None
    required: bool | None = None
    deprecated: bool | None = None
    allowEmptyValue: bool | None = None
    style: str | None = None
    explode: bool | None = None
    allowReserved: bool | None = None
    schema_: SchemaReferenceType | None = Field(alias="schema")
    example: Any | None = None
    examples: dict[str, ExampleReferenceType] | None = None
    content: dict[str, "MediaTypeObject"] | None = None  # noqa: UP037
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#parameter-object"
    )

    _in_valid = mv.if_then(
        conditions={"in_": mv.UNKNOWN},
        consequences={"in_": ["query", "header", "path", "cookie"]},
    )
    _path_location_is_required = mv.if_then(
        conditions={"in_": "path"}, consequences={"required": True}
    )
    _empty_value_only_with_query = mv.if_then(
        conditions={"allowEmptyValue": mv.UNKNOWN},
        consequences={"in_": PARAMETER_STYLES ^ {"query"}},
    )
    _style_is_valid = mv.if_then(
        conditions={"style": mv.UNKNOWN}, consequences={"style": list(PARAMETER_STYLES)}
    )
    _reserved_only_with_query = mv.if_then(
        conditions={"allowReserved": mv.UNKNOWN},
        consequences={"in_": PARAMETER_STYLES ^ {"query"}},
    )
    _disallowed_if_schema = mv.if_then(
        conditions={"schema_": mv.UNKNOWN}, consequences={"content": None}
    )
    _disallowed_if_content = mv.if_then(
        conditions={"content": mv.UNKNOWN},
        consequences={
            "content": None,
            "style": None,
            "explode": None,
            "allowReserved": None,
            "schema_": None,
        },
    )


@specification_extensions("x-")
class RequestBodyObject(GenericObject):
    """
    Validates the OpenAPI Specification request body object - §4.8.13
    """

    description: CommonMark | str | None = None
    content: dict[str, MediaTypeObject]
    required: bool | None = False
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#request-body-object"
    )


@specification_extensions("x-")
class MediaTypeObject(GenericObject):
    """
    Validates the OpenAPI Specification media type object - §4.8.14
    """

    schema_: SchemaReferenceType | None = Field(alias="schema", default=None)
    # FIXME: Define example
    example: Any | None = None
    examples: dict[str, ExampleReferenceType] | None = None
    encoding: EncodingObject | None = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#media-type-object"
    )


@specification_extensions("x-")
class EncodingObject(GenericObject):
    """
    Validates the OpenAPI Specification media type object - §4.8.15
    """

    contentType: str | None = None
    headers: dict[str, HeaderReferenceType] | None = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#encoding object-object"
    )

    @field_validator("contentType", mode="after")
    @classmethod
    def check_content_type(cls, value: str) -> str:
        """
        contentType is a comma-separated list of media types.
        Check that they are all valid

        raises: ValueError
        """

        for media_type in value.split(","):
            MediaType(media_type.strip())

        return value


type _ResponsesObjectReturnType = dict[str, ResponseReferenceType]


@specification_extensions(".*")
class ResponsesObject(GenericObject):
    """
    Validates the OpenAPI Specification responses object - §4.8.16
    """

    model_config: ClassVar[ConfigDict] = ConfigDict(
        extra="allow",
    )

    default: ResponseReferenceType | None = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#responses-object"
    )

    @classmethod
    def _choose_model(
        cls, value: Any, field_name: str
    ) -> ReferenceObject | ResponseObject:
        """
        Choose the model to use for validation based on the type of value.

        Args:
            value: The value to validate.

        Returns:
            The model class to use for validation.
        """

        message = f"{field_name} must be a ResponseObject or ReferenceObject, got {type(value)}"  # noqa: E501

        try:
            return ResponseObject.model_validate(value)
        except ValidationError:
            try:
                return ReferenceObject.model_validate(value)
            except ValidationError as e:
                raise ValueError(message, ResponsesObject._reference_uri) from e

    @model_validator(mode="before")
    @classmethod
    def validate_all_fields(cls, data: dict[str, Any]) -> _ResponsesObjectReturnType:
        """
        Validates the responses object.
        """

        validated_data: _ResponsesObjectReturnType = {}

        for field_name, value in data.items():
            # If the value is a specification extension, allow it
            if field_name.startswith("x-"):
                validated_data[field_name] = value
                continue

            # If the value is the fixed field, "default", allow it
            if field_name == "default":
                if isinstance(value, dict):
                    validated_data[field_name] = ResponsesObject._choose_model(
                        value, field_name
                    )
                continue

            # Otherwise, if the field appears like a valid HTTP status code or a range
            if re.match(r"^[1-5]([0-9]{2}|XX)+$", str(field_name)):
                # Double check and raise a value error if not
                HTTPStatusCode(field_name)

                # and validate as a ResponseObject or ReferenceObject
                validated_data[field_name] = ResponsesObject._choose_model(
                    value, field_name
                )

                continue

            # If the field is not a valid HTTP status code or "default"
            raise ValueError(f"Invalid type for numeric field '{field_name}'")

        return validated_data


@specification_extensions("x-")
class ResponseObject(GenericObject):
    """
    Validates the OpenAPI Specification response object - §4.8.17
    """

    description: str | CommonMark
    headers: dict[str, HeaderReferenceType] | None = None
    content: dict[str, MediaTypeObject] | None = None
    links: dict[str, LinkReferenceType] | None = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#response-object"
    )


@specification_extensions("x-")
class CallbackObject(GenericObject):
    """
    Validates the OpenAPI Specification callback object - §4.8.18
    """

    model_config: ClassVar[ConfigDict] = ConfigDict(extra="allow")

    # The keys are runtime expressions that resolve to a URL
    # The values are Response Objects or Reference Objects
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#callback-object"
    )

    @model_validator(mode="before")
    @classmethod
    def validate_all_fields(cls, data: dict[str, Any]) -> dict[str, PathItemObject]:
        """
        Validates the callback object.
        """

        validated_data: dict[str, PathItemObject] = {}

        # Everything after a { but before a } should be runtime expression
        pattern: str = r"\{([^}]+)\}"

        for field_name, value in data.items():
            # If the value is a specification extension, allow it
            if field_name.startswith("x-"):
                validated_data[field_name] = PathItemObject.model_validate(value)
                continue

            # Either the field name is a runtime expression, so test this:
            try:
                RuntimeExpression(field_name)
                validated_data[field_name] = PathItemObject.model_validate(value)
                continue
            except AmatiValueError:
                pass

            # Or, the field name is a runtime expression embedded in a string
            # value per https://spec.openapis.org/oas/latest.html#examples-0
            matches = re.findall(pattern, field_name)

            for match in matches:
                try:
                    RuntimeExpression(match)
                except AmatiValueError as e:
                    raise AmatiValueError(
                        f"Invalid runtime expression '{match}' in field '{field_name}'",
                        CallbackObject._reference_uri,
                    ) from e

            if matches:
                validated_data[field_name] = PathItemObject.model_validate(value)
            else:
                # If the field does not contain a valid runtime expression
                raise ValueError(f"Invalid type for numeric field '{field_name}'")

        return validated_data


@specification_extensions("x-")
class TagObject(GenericObject):
    """
    Validates the OpenAPI Specification tag object - §4.8.22
    """

    name: str
    description: str | CommonMark | None = None
    externalDocs: ExternalDocumentationObject | None = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#tag-object"
    )


@specification_extensions("x-")
class LinkObject(GenericObject):
    """
    Validates the OpenAPI Specification link object - §4.8.20
    """

    operationRef: URI | None = None
    operationId: str | None = None
    parameters: dict[str, RuntimeExpression | JSONValue] | None = None
    requestBody: JSONValue | RuntimeExpression | None = None
    description: str | CommonMark | None = None
    server: ServerObject | None = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#link-object"
    )

    _not_operationref_and_operationid = mv.only_one_of(
        fields=["operationRef", "operationId"]
    )


@specification_extensions("x-")
class HeaderObject(GenericObject):
    """
    Validates the OpenAPI Specification link object - §4.8.20
    """

    # Common schema/content fields
    description: str | CommonMark | None = None
    required: bool | None = Field(default=False)
    deprecated: bool | None = Field(default=False)

    # Schema fields
    style: str | None = Field(default="simple")
    explode: bool | None = Field(default=False)
    schema_: SchemaReferenceType | None = Field(alias="schema", default=None)
    example: JSONValue | None = None
    examples: dict[str, ExampleReferenceType] | None = None

    # Content fields
    content: dict[str, MediaTypeObject] | None = None

    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#link-object"
    )

    _not_schema_and_content = mv.only_one_of(["schema_", "content"])


@specification_extensions("x-")
class XMLObject(GenericObject):
    """
    Validates the OpenAPI Specification object - §4.8.26
    """

    name: str | None = None
    namespace: URI | None = None
    prefix: str | None = None
    attribute: bool | None = Field(default=False)
    wrapped: bool | None = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#xml-object"
    )

    @field_validator("namespace", mode="after")
    @classmethod
    def _validate_namespace(cls, value: URI) -> URI:
        """
        Validates that the namespace is not a relative URI.
        """
        if value.type == URIType.RELATIVE:
            message = "XML namespace {value} cannot be a relative URI"
            Logger.log(
                {
                    "msg": message,
                    "type": "value_error",
                    "loc": (cls.__name__,),
                    "input": value,
                    "url": cls._reference_uri,
                }
            )

        return value


class SchemaObject(GenericObject):
    """
    Schema Object as per OAS 3.1.1 specification (section 4.8.24)

    This model defines only the OpenAPI-specific fields explicitly.
    Standard JSON Schema fields are allowed via the 'extra' config
    and validated through jsonschema.
    """

    model_config: ClassVar[ConfigDict] = ConfigDict(
        populate_by_name=True,
        extra="allow",  # Allow all standard JSON Schema fields
    )

    # OpenAPI-specific fields not in standard JSON Schema
    nullable: bool | None = None  # OAS 3.0 style nullable flag
    discriminator: DiscriminatorObject | None = None  # Polymorphism support
    readOnly: bool | None = None  # Declares property as read-only for requests
    writeOnly: bool | None = None  # Declares property as write-only for responses
    xml: XMLObject | None = None  # XML metadata
    externalDocs: ExternalDocumentationObject | None = None  # External documentation
    example: Any | None = None  # Example of schema
    examples: list[Any] | None = None  # Examples of schema (OAS 3.1)
    deprecated: bool | None = None  # Specifies schema is deprecated

    # JSON Schema fields that need special handling in OAS context
    ref: str | None = Field(default=None, alias="$ref")  # Reference to another schema

    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#schema-object"
    )

    @model_validator(mode="after")
    def validate_schema(self):
        """
        Use jsonschema to validate the model as a valid JSON Schema
        """
        schema_dict = self.model_dump(exclude_none=True, by_alias=True)

        # Handle OAS 3.1 specific validations

        # 1. Convert nullable to type array with null if needed
        if schema_dict.get("nullable") is True and "type" in schema_dict:
            type_val = schema_dict["type"]
            if isinstance(type_val, str) and type_val != "null":
                schema_dict["type"] = [type_val, "null"]
            elif isinstance(type_val, list) and "null" not in type_val:
                schema_dict["type"] = [*type_val, ["null"]]

        # 2. Validate the schema structure using jsonschema's meta-schema
        # Get the right validator based on the declared $schema or default
        # to Draft 2020-12
        schema_version = schema_dict.get(
            "$schema", "https://json-schema.org/draft/2020-12/schema"
        )
        try:
            validator_cls: JSONSchemaValidator = validator_for(  # type: ignore
                {"$schema": schema_version}
            )
            meta_schema: JSON = validator_cls.META_SCHEMA  # type: ignore

            # This will validate the structure conforms to JSON Schema
            validator_cls(meta_schema).validate(schema_dict)  # type: ignore
        except JSONVSchemeValidationError as e:
            Logger.log(
                {
                    "msg": f"Invalid JSON Schema: {e.message}",
                    "type": "value_error",
                    "loc": (self.__class__.__name__,),
                    "input": schema_dict,
                    "url": self._reference_uri,
                }
            )

        return self


OAUTH_FLOW_TYPES: set[str] = {
    "implicit",
    "authorizationCode",
    "clientCredentials",
    "password",
}


@specification_extensions("x-")
class OAuthFlowObject(GenericObject):
    """
    Validates the OpenAPI OAuth Flow object - §4.8.29
    """

    type: str | None = None
    authorizationUrl: URI | None = None
    tokenUrl: URI | None = None
    refreshUrl: URI | None = None
    scopes: dict[str, str] = Field(default={})
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#oauth-flow-object"
    )

    _implicit_has_authorization_url = mv.if_then(
        conditions={"type": "implicit"},
        consequences={"authorizationUrl": mv.UNKNOWN},
    )

    _token_url_not_implicit = mv.if_then(
        conditions={"tokenUrl": mv.UNKNOWN},
        consequences={"type": OAUTH_FLOW_TYPES ^ {"implicit"}},
    )

    _authorization_code_has_urls = mv.if_then(
        conditions={"type": "authorizationCode"},
        consequences={"authorizationUrl": mv.UNKNOWN, "tokenUrl": mv.UNKNOWN},
    )

    _authorization_url_not_credentials_password = mv.if_then(
        conditions={"authorizationUrl": mv.UNKNOWN},
        consequences={"type": OAUTH_FLOW_TYPES ^ {"clientCredentials", "password"}},
    )

    _client_credentials_has_token = mv.if_then(
        conditions={"type": "clientCredentials"},
        consequences={"tokenUrl": mv.UNKNOWN},
    )
    _password_has_token = mv.if_then(
        conditions={"type": "password"}, consequences={"tokenUrl": mv.UNKNOWN}
    )

    @model_serializer(mode="wrap", when_used="json")
    def serialize_model(
        self, handler: SerializerFunctionWrapHandler
    ) -> dict[str, object]:
        """
        Serializes the type field for JSON output only.
        The Python output is used inside model_validators, so
        a standard Field(exclude=True) cannot be used.
        """

        serialized = handler(self)
        serialized.pop("type", None)
        return serialized


@specification_extensions("-x")
class OAuthFlowsObject(GenericObject):
    """
    Validates the OpenAPI OAuth Flows object - §4.8.28

    SPECFIX: Not all of these should be optional as an OAuth2 workflow
    without any credentials will not do anything.
    """

    implicit: OAuthFlowObject | None = None
    password: OAuthFlowObject | None = None
    clientCredentials: OAuthFlowObject | None = None
    authorizationCode: OAuthFlowObject | None = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#oauth-flow-object"
    )

    @model_validator(mode="before")
    @classmethod
    def _push_down_type(cls, data: Any) -> Any:
        """
        Adds the type of OAuth2 flow, e.g. implicit, password, to the child
        OAuthFlowObject so that additional validation can be done on this object.
        """

        for field, value in data.items():
            if isinstance(value, OAuthFlowObject):
                raise NotImplementedError("Must pass a dict")
            data[field]["type"] = field

        return data


class SecuritySchemeObject(GenericObject):
    """
    Validates the OpenAPI Security Scheme object - §4.8.27
    """

    type: str
    description: str | CommonMark | None = None
    name: str | None = None
    in_: str | None = Field(default=None, alias="in")
    scheme: HTTPAuthenticationScheme | None = None
    bearerFormat: str | None = None
    flows: OAuthFlowsObject | None = None
    openIdConnectUrl: URI | None = None

    _SECURITY_SCHEME_TYPES: ClassVar[set[str]] = {
        "apiKey",
        "http",
        "oauth2",
        "openIdConnect",
    }

    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#security-scheme-object-0"
    )

    _type_in_enum = mv.if_then(
        conditions={"type": mv.UNKNOWN}, consequences={"type": _SECURITY_SCHEME_TYPES}
    )

    _apikey_has_name_and_in = mv.if_then(
        conditions={"type": "apiKey"},
        consequences={"name": mv.UNKNOWN, "in_": ("query", "header", "cookie")},
    )

    _http_has_scheme = mv.if_then(
        conditions={"type": "http"}, consequences={"scheme": mv.UNKNOWN}
    )

    _oauth2_has_flows = mv.if_then(
        conditions={"type": "oauth2"}, consequences={"flows": mv.UNKNOWN}
    )

    _open_id_connect_has_url = mv.if_then(
        conditions={"type": "openIdConnect"},
        consequences={"openIdConnectUrl": mv.UNKNOWN},
    )

    _flows_not_oauth2 = mv.if_then(
        conditions={"flows": None},
        consequences={"type": _SECURITY_SCHEME_TYPES ^ {"oauth2"}},
    )


type _Requirement = dict[str, list[str]]


# NB This is implemented as a RootModel as there are no pre-defined field names.
class SecurityRequirementObject(RootModel[list[_Requirement] | _Requirement]):
    """
    Validates the OpenAPI Specification security requirement object - §4.8.30:
    """

    # FIXME: The name must be a valid Security Scheme - need to use post-processing
    # FIXME If the security scheme is of type "oauth2" or "openIdConnect", then the
    # value must be a list For other security scheme types, the array MAY contain a
    # list of role names which are required for the execution
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/3.0.4.html#security-requirement-object"
    )


@specification_extensions("x-")
class OperationObject(GenericObject):
    """Validates the OpenAPI Specification operation object - §4.8.10"""

    tags: list[str] | None = None
    summary: str | None = None
    description: str | CommonMark | None = None
    externalDocs: ExternalDocumentationObject | None = None
    operationId: str | None = None
    parameters: list[ParameterReferenceType] | None = None
    requestBody: RequestBodyReferenceType | None = None
    responses: ResponsesObject
    callbacks: dict[str, CallbackReferenceType] | None = None
    deprecated: bool | None = False
    security: list[SecurityRequirementObject] | None = None
    servers: list[ServerObject] | None = None

    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#operation-object"
    )


@specification_extensions("x-")
class PathItemObject(GenericObject):
    """Validates the OpenAPI Specification path item object - §4.8.9"""

    ref_: URI | None = Field(alias="$ref", default=None)
    summary: str | None = None
    description: str | CommonMark | None = None
    get: OperationObject | None = None
    put: OperationObject | None = None
    post: OperationObject | None = None
    delete: OperationObject | None = None
    options: OperationObject | None = None
    head: OperationObject | None = None
    patch: OperationObject | None = None
    trace: OperationObject | None = None
    servers: list[ServerObject] | None = None
    parameters: list[ParameterReferenceType] | None = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#path-item-object"
    )


@specification_extensions("x-")
class ComponentsObject(GenericObject):
    """
    Validates the OpenAPI Specification components object - §4.8.7
    """

    schemas: dict[str, SchemaObject] | None = None
    responses: dict[str, ResponseReferenceType] | None = None
    parameters: dict[str, ParameterReferenceType] | None = None
    examples: dict[str, ExampleReferenceType] | None = None
    requestBodies: dict[str, RequestBodyReferenceType] | None = None
    headers: dict[str, HeaderReferenceType] | None = None
    securitySchemes: dict[str, SecuritySchemeReferenceType] | None = None
    links: dict[str, LinkReferenceType] | None = None
    callbacks: dict[str, CallbackReferenceType] | None = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#components-object"
    )

    @model_validator(mode="before")
    @classmethod
    def validate_all_fields(
        cls, data: dict[str, dict[str, Any]]
    ) -> dict[str, dict[str, Any]]:
        """
        Validates the components object.

        Args:
            data: The data to validate.

        Returns:
            The validated components object.
        """

        pattern: str = r"^[a-zA-Z0-9\.\-_]+$"

        # Validate each field in the components object
        for field_name, value in data.items():
            if field_name.startswith("x-"):
                continue

            if not isinstance(value, dict):  # type: ignore
                raise ValueError(
                    f"Invalid type for '{field_name}': expected dict, got {type(value)}"
                )

            for key in value:
                if not re.match(pattern, key):
                    raise ValueError(
                        f"Invalid key '{key}' in '{field_name}': must match pattern {pattern}"  # noqa: E501
                    )

        return data


@specification_extensions("x-")
class OpenAPIObject(GenericObject):
    """
    Validates the OpenAPI Specification object - §4.1
    """

    openapi: OpenAPI
    info: InfoObject
    servers: list[ServerObject] | None = Field(default=[ServerObject(url=URI("/"))])
    paths: PathsObject
    components: ComponentsObject | None = None
    security: list[SecurityRequirementObject] | None = None
    tags: list[TagObject] | None = None
    externalDocs: ExternalDocumentationObject | None = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/3.0.4.html#openapi-object"
    )



"""
Validates IANA HTTP status codes,

Note that the codes ^[1-5]XX$ are not valid HTTP status codes,
but are in common usage. They can be accessed separately via HTTPStatusCodeX,
or the numeric codes can be accessed via HTTPStatusCodeN.
"""

import re
from typing import Self, cast

from amati import get
from amati.exceptions import AmatiValueError
from amati.fields import Str as _Str

reference_uri = (
    "https://www.iana.org/assignments/http-status-codes/http-status-codes.xhtml"
)

HTTP_STATUS_CODES = cast(dict[str, str], get("http_status_code"))


class HTTPStatusCode(_Str):
    """
    A class representing an HTTP status code as defined in RFC 7231 and related RFCs.

    This class validates and provides information about HTTP status codes, including
    whether they are registered with IANA, assigned for use, or represent a status code
    range (e.g., 4XX). It inherits from _Str to maintain compatibility with string
    operations while adding HTTP-specific validation and metadata.

    Attributes:
        description (Optional[str]): The official description of the status code,
            if registered.
        is_registered (bool): Whether the status code is registered with IANA.
        is_assigned (bool): Whether the status code is assigned for use
            (not marked as 'Unassigned').
        is_range (bool): Whether the status code represents a range (e.g., "2XX").
    """

    description: str | None = None
    is_registered: bool = False
    is_assigned: bool = False
    is_range: bool = False
    _pattern = re.compile(r"^[1-5]XX$")

    def __init__(self: Self, value: str):
        """
        Initialize an HTTPStatusCode object from a string or integer value.

        Validates the input against known HTTP status codes and status code ranges.
        Sets attributes indicating whether the code is registered, assigned, and/or
        represents a range.

        NB: OAS 3.1.1 states that:
        > This field MUST be enclosed in quotation marks (for example, “200”)
        > for compatibility between JSON and YAML.

        Args:
            value (str | int): A string or integer representing an HTTP status code
                (e.g., "200", 404) or status code range (e.g., "4XX").

        Raises:
            AmatiValueError: If the provided value is not a valid HTTP status code or
                status code range.
        """

        # Type-hinting that something should be a string is not enough
        # double check that a string will be returned in the models.
        if not isinstance(value, str):  # type: ignore
            type_ = type(value).__name__
            raise AmatiValueError(
                f"{value} of type {type_} cannot be a valid HTTP Status code."
            )

        candidate = str(value)

        if candidate in HTTP_STATUS_CODES:
            self.is_registered = True
            self.description = HTTP_STATUS_CODES[candidate]
        elif self._pattern.match(candidate):
            self.is_range = True
        else:
            raise AmatiValueError(
                f"{value} is not a valid HTTP Status Code", reference_uri
            )

        if self.description != "Unassigned":
            self.is_assigned = True



"""
Tests amati/fields/http_status_codes.py
"""

from typing import Any

import pytest
from hypothesis import given
from hypothesis.strategies import integers, sampled_from

from amati import AmatiValueError
from amati.fields.http_status_codes import HTTP_STATUS_CODES, HTTPStatusCode
from tests import strategies

REGISTERED_HTTP_STATUS_CODES = list(HTTP_STATUS_CODES.keys())
HTTP_STATUS_CODE_RANGES = ["1XX", "2XX", "3XX", "4XX", "5XX"]


@given(sampled_from(REGISTERED_HTTP_STATUS_CODES))
def test_registered_status_code(value: str):
    result = HTTPStatusCode(value)

    if HTTP_STATUS_CODES[value] == "Unassigned":
        assert result.is_assigned is False

    assert result.is_registered is True


@given(integers(max_value=99))
def test_invalid_status_code_below_range(value: str):
    with pytest.raises(AmatiValueError):
        HTTPStatusCode(value)


@given(integers(min_value=600))
def test_invalid_status_code_above_range(value: str):
    with pytest.raises(AmatiValueError):
        HTTPStatusCode(value)


@given(sampled_from(HTTP_STATUS_CODE_RANGES))
def test_status_code_range(value: str):
    result = HTTPStatusCode(value)

    assert result.is_registered is False
    assert result.is_range is True


@given(
    strategies.everything_except(int).filter(
        lambda x: x not in REGISTERED_HTTP_STATUS_CODES
    )
)
def test_everything_except_integers(value: Any):
    with pytest.raises(AmatiValueError):
        HTTPStatusCode(value)  # type: ignore



"""Tests for amati.model_validators.if_then"""

from typing import ClassVar

import pytest
from hypothesis import given
from hypothesis import strategies as st
from pydantic import BaseModel

from amati._logging import Logger
from amati.model_validators import UNKNOWN, if_then


class ModelNoConditions(BaseModel):
    field: str = ""
    x = if_then()
    _reference_uri: ClassVar[str] = "https://example.com"


class ModelWithConditions(BaseModel):
    role: str = ""
    permission: bool = False
    _if_then = if_then(conditions={"role": "admin"}, consequences={"permission": True})
    _reference_uri: ClassVar[str] = "https://example.com"


class ModelWithMultipleConditions(BaseModel):
    role: str = ""
    department: str = ""
    can_approve: bool = False
    _if_then = if_then(
        conditions={"role": "manager", "department": "finance"},
        consequences={"can_approve": True},
    )
    _reference_uri: ClassVar[str] = "https://example.com"


# Add new test class after existing classes
class ModelWithUnknownConditions(BaseModel):
    role: str = ""
    permission: bool = False
    notify: bool = False
    _if_then = if_then(
        conditions={"role": "admin"},
        consequences={"permission": True, "notify": UNKNOWN},
    )
    _reference_uri: ClassVar[str] = "https://example.com"


class ModelWithMultipleUnknownConditions(BaseModel):
    role: str = ""
    department: str = ""
    can_approve: bool = False
    can_edit: bool = False
    _if_then = if_then(
        conditions={"role": "manager", "department": "finance"},
        consequences={"can_approve": True, "can_edit": UNKNOWN},
    )
    _reference_uri: ClassVar[str] = "https://example.com"


class ModelWithIterableConditions(BaseModel):
    role: str = ""
    status: str = ""
    _if_then = if_then(
        conditions={"role": "admin"}, consequences={"status": ["active", "pending"]}
    )
    _reference_uri: ClassVar[str] = "https://example.com"


class ModelWithMultipleIterableConditions(BaseModel):
    role: str = ""
    department: str = ""
    status: str = ""
    level: int = 0
    _if_then = if_then(
        conditions={"role": "manager", "department": "finance"},
        consequences={"status": ["active", "pending"], "level": [1, 2, 3]},
    )
    _reference_uri: ClassVar[str] = "https://example.com"


def test_missing_conditions_consequences():
    """Test that validator raises error when conditions/consequences missing."""
    with pytest.raises(ValueError) as exc:
        ModelNoConditions()
    assert "A condition and a consequence must be present" in str(exc.value)


def test_conditions_not_met():
    """Test that validation passes when conditions are not met."""
    with Logger.context():
        model = ModelWithConditions(role="user", permission=False)
        assert not Logger.logs
        assert model.role == "user"
        assert model.permission is False


def test_conditions_met_valid():
    """Test that validation passes when conditions are met and consequences valid."""
    with Logger.context():
        model = ModelWithConditions(role="admin", permission=True)
        assert not Logger.logs
        assert model.role == "admin"
        assert model.permission is True


def test_conditions_met_invalid():
    """Test that validation fails when conditions met but consequences invalid."""
    with Logger.context():
        ModelWithConditions(role="admin", permission=False)
        assert len(Logger.logs) == 1
        assert "Expected permission to be True found False" in Logger.logs[0]["msg"]


@given(role=st.sampled_from(["admin", "user", ""]), permission=st.booleans())
def test_property_based(role: str, permission: bool):
    with Logger.context():
        ModelWithConditions(role=role, permission=permission)

        if role == "admin" and not permission:
            assert len(Logger.logs) == 1
            assert "Expected permission to be True" in Logger.logs[0]["msg"]
        else:
            assert len(Logger.logs) == 0


def test_multiple_conditions_all_met_valid():
    """Test validation passes when all conditions are met and consequences valid."""
    with Logger.context():
        model = ModelWithMultipleConditions(
            role="manager", department="finance", can_approve=True
        )
        assert not Logger.logs
        assert model.role == "manager"
        assert model.department == "finance"
        assert model.can_approve is True


def test_multiple_conditions_all_met_invalid():
    """Test validation fails when all conditions met but consequences invalid."""
    with Logger.context():
        ModelWithMultipleConditions(
            role="manager", department="finance", can_approve=False
        )
        assert len(Logger.logs) == 1
        assert "Expected can_approve to be True found False" in Logger.logs[0]["msg"]


def test_multiple_conditions_partial_met():
    """Test validation passes when only some conditions are met."""
    with Logger.context():
        cases: list[dict[str, str | bool]] = [
            {"role": "manager", "department": "hr", "can_approve": False},
            {"role": "employee", "department": "finance", "can_approve": False},
            {"role": "employee", "department": "hr", "can_approve": False},
        ]

        for case in cases:
            model = ModelWithMultipleConditions(**case)  # type: ignore
            assert not Logger.logs
            assert model.role == case["role"]
            assert model.department == case["department"]
            assert model.can_approve is False


@given(
    role=st.sampled_from(["manager", "employee", "admin"]),
    department=st.sampled_from(["finance", "hr", "it"]),
    can_approve=st.booleans(),
)
def test_multiple_conditions_property_based(
    role: str, department: str, can_approve: bool
):
    """Property-based test for multiple conditions."""
    with Logger.context():
        ModelWithMultipleConditions(
            role=role, department=department, can_approve=can_approve
        )

        if role == "manager" and department == "finance" and not can_approve:
            assert len(Logger.logs) == 1
            assert "Expected can_approve to be True" in Logger.logs[0]["msg"]
        else:
            assert len(Logger.logs) == 0


# Add new test functions before property-based tests
def test_unknown_consequence_any_value():
    """Test that UNKNOWN consequence allows any value when conditions are met."""
    with Logger.context():
        # Should pass with notify=True
        model = ModelWithUnknownConditions(role="admin", permission=True, notify=True)
        assert not Logger.logs
        assert model.notify is True

        # Should also pass with notify=False
        model = ModelWithUnknownConditions(role="admin", permission=True, notify=False)
        assert not Logger.logs
        assert model.notify is False


def test_unknown_consequence_conditions_not_met():
    """Test that UNKNOWN consequence is ignored when conditions are not met."""
    with Logger.context():
        model = ModelWithUnknownConditions(role="user", permission=False, notify=False)
        assert not Logger.logs
        assert model.notify is False


def test_multiple_conditions_unknown_consequence():
    """Test multiple conditions with UNKNOWN consequence."""
    with Logger.context():
        # Should pass with can_edit=True
        model = ModelWithMultipleUnknownConditions(
            role="manager", department="finance", can_approve=True, can_edit=True
        )
        assert not Logger.logs
        assert model.can_edit is True

        # Should also pass with can_edit=False
        model = ModelWithMultipleUnknownConditions(
            role="manager", department="finance", can_approve=True, can_edit=False
        )
        assert not Logger.logs
        assert model.can_edit is False


# Add to existing property-based test or create new one
@given(
    role=st.sampled_from(["admin", "user"]),
    permission=st.booleans(),
    notify=st.booleans(),
)
def test_unknown_property_based(role: str, permission: bool, notify: bool):
    """Property-based test for UNKNOWN consequences."""
    with Logger.context():
        model = ModelWithUnknownConditions(
            role=role, permission=permission, notify=notify
        )

        if role == "admin" and not permission:
            assert len(Logger.logs) == 1
            assert "Expected permission to be True" in Logger.logs[0]["msg"]
        else:
            assert len(Logger.logs) == 0
            if role == "admin":
                # When conditions are met, notify can be any value
                assert model.notify == notify


# Add after existing test functions
def test_iterable_consequence_valid():
    """Test that validation passes when value is in iterable consequence."""
    with Logger.context():
        # Should pass with status in allowed values
        model = ModelWithIterableConditions(role="admin", status="active")
        assert not Logger.logs
        assert model.status == "active"

        model = ModelWithIterableConditions(role="admin", status="pending")
        assert not Logger.logs
        assert model.status == "pending"


def test_iterable_consequence_invalid():
    """Test that validation fails when value not in iterable consequence."""
    with Logger.context():
        ModelWithIterableConditions(role="admin", status="inactive")
        assert len(Logger.logs) == 1
        assert (
            "Expected status to be in ['active', 'pending'] found inactive"
            in Logger.logs[0]["msg"]
        )


def test_iterable_consequence_conditions_not_met():
    """Test that validation passes when conditions not met, regardless of value."""
    with Logger.context():
        model = ModelWithIterableConditions(role="user", status="whatever")
        assert not Logger.logs
        assert model.status == "whatever"


def test_multiple_iterable_consequences_valid():
    """Test that validation passes when all values are in their respective iterables."""
    with Logger.context():
        model = ModelWithMultipleIterableConditions(
            role="manager", department="finance", status="active", level=2
        )
        assert not Logger.logs
        assert model.status == "active"
        assert model.level == 2  # noqa: PLR2004


def test_multiple_iterable_consequences_partial_invalid():
    """Test that validation fails when any value is not in its iterable."""
    with Logger.context():
        # Invalid status
        ModelWithMultipleIterableConditions(
            role="manager", department="finance", status="inactive", level=2
        )
        assert len(Logger.logs) == 1
        Logger.logs.clear()

        # Invalid level
        ModelWithMultipleIterableConditions(
            role="manager", department="finance", status="active", level=4
        )
        assert len(Logger.logs) == 1


@given(
    role=st.sampled_from(["admin", "user"]),
    status=st.sampled_from(["active", "pending", "inactive", "suspended"]),
)
def test_iterable_property_based(role: str, status: str):
    """Property-based test for iterable consequences."""
    with Logger.context():
        model = ModelWithIterableConditions(role=role, status=status)

        if role == "admin" and status not in ["active", "pending"]:
            assert len(Logger.logs) == 1
            assert (
                f"Expected status to be in ['active', 'pending'] found {status}"
                in Logger.logs[0]["msg"]
            )
        else:
            assert len(Logger.logs) == 0
            if role == "admin":
                assert model.status in ["active", "pending"]


def test_if_then_none_unknown_condition():
    """Test if_then when a field is None but condition expects UNKNOWN."""

    class ConditionalModel(BaseModel):
        status: str | None = None
        result: bool = False
        _if_then = if_then(
            conditions={"status": UNKNOWN}, consequences={"result": True}
        )
        _reference_uri: ClassVar[str] = "https://example.com"

    with Logger.context():
        ConditionalModel()
        assert not Logger.logs  # Should early return due to None vs UNKNOWN check


def test_if_then_unfulfilled_condition():
    """Test if_then when a field is None but condition expects UNKNOWN."""

    class ConditionalModel(BaseModel):
        status: str = "Present"
        result: bool = False
        _if_then = if_then(
            conditions={"status": "Not Present"}, consequences={"result": True}
        )
        _reference_uri: ClassVar[str] = "https://example.com"

    with Logger.context():
        ConditionalModel()
        assert not Logger.logs



"""
Logging utilities for Amati.
"""

from collections.abc import Generator
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Any, ClassVar, NotRequired, TypedDict

type LogType = Exception | Warning


@dataclass(frozen=True)
class Log(TypedDict):
    type: str
    loc: NotRequired[tuple[int | str, ...]]
    msg: str
    input: NotRequired[Any]
    url: NotRequired[str]


class Logger:
    """A simple class-level logger for collecting Log objects.

    This class provides methods for appending logs and managing
    a logging context that automatically clears the logs.
    """

    logs: ClassVar[list[Log]] = []

    @classmethod
    def log(cls, message: Log) -> None:
        """Add a new message to the logs list.

        Args:
            message: A Log object containing the message to be logged.

        Returns:
            The current list of logs after adding the new message.
        """
        cls.logs.append(message)

    @classmethod
    @contextmanager
    def context(cls) -> Generator[list[Log]]:
        """Create a context manager for handling logs.

        Yields:
            The current list of logs.

        Notes:
            Automatically clears the logs when exiting the context.
        """
        try:
            yield cls.logs
        finally:
            cls.logs.clear()
